@@ -78,7 +78,9 @@ if ($config{'mail_system'} == 1) {
 	print LOCALS $_[0]->{'dom'},"\n";
 	close(LOCALS);
 	&unlock_file($cwfile);
-	&sendmail::restart_sendmail();
+	if (!$no_restart_mail) {
+		&sendmail::restart_sendmail();
+		}
 	}
 elsif ($config{'mail_system'} == 0) {
 	# Add a special postfix virtual entry just for the domain
@@ -98,7 +100,9 @@ elsif ($config{'mail_system'} == 2) {
 	local $virtmap = { 'domain' => $_[0]->{'dom'},
 			   'prepend' => $_[0]->{'group'} };
 	&qmailadmin::create_virt($virtmap);
-	&qmailadmin::restart_qmail();
+	if (!$no_restart_mail) {
+		&qmailadmin::restart_qmail();
+		}
 	}
 &$second_print($text{'setup_done'});
 }
@@ -124,7 +128,9 @@ if ($config{'mail_system'} == 1) {
 	print LOCALS @locals;
 	close(LOCALS);
 	&unlock_file($cwfile);
-	&sendmail::restart_sendmail();
+	if (!$no_restart_mail) {
+		&sendmail::restart_sendmail();
+		}
 	}
 elsif ($config{'mail_system'} == 0) {
 	# Delete the special postfix virtuser
@@ -141,8 +147,10 @@ elsif ($config{'mail_system'} == 0) {
 		splice(@md, $idx, 1);
 		&postfix::set_current_value("mydestination", join(", ", @md));
 		&unlock_file($postfix::config{'postfix_config_file'});
-		&shutdown_mail_server();
-		&startup_mail_server();
+		if (!$no_restart_mail) {
+			&shutdown_mail_server();
+			&startup_mail_server();
+			}
 		}
 	}
 elsif ($config{'mail_system'} == 2) {
@@ -158,7 +166,9 @@ elsif ($config{'mail_system'} == 2) {
 	local ($virtmap) = grep { $_->{'domain'} eq $_[0]->{'dom'} &&
 				  !$_->{'user'} } &qmailadmin::list_virts();
 	&qmailadmin::delete_virt($virtmap) if ($virtmap);
-	&qmailadmin::restart_qmail();
+	if (!$no_restart_mail) {
+		&qmailadmin::restart_qmail();
+		}
 	}
 &$second_print($text{'setup_done'});
 }
@@ -167,9 +177,45 @@ elsif ($config{'mail_system'} == 2) {
 # Deal with a change in domain name
 sub modify_mail
 {
+if ($_[0]->{'home'} ne $_[1]->{'home'}) {
+	# Need to update the home directory of all mail users .. but only
+	# in the Unix object, as their files will have already been moved
+	# as part of the domain's directory
+	&$first_print($text{'save_mailhomes'});
+	local $u;
+	foreach $u (&list_domain_users($_[0], 1)) {
+		local %oldu = %$u;
+		$u->{'home'} =~ s/$_[1]->{'home'}/$_[0]->{'home'}/;
+		&modify_user($u, \%oldu, $_[0], 1);
+		}
+	&$second_print($text{'setup_done'});
+	}
 if ($_[0]->{'dom'} ne $_[1]->{'dom'}) {
+	# Delete the old mail domain and add the new
+	local $no_restart_mail = 1;
 	&delete_mail($_[1]);
 	&setup_mail($_[0]);
+	&require_mail();
+	if (&is_mail_running()) {
+		if ($config{'mail_system'} == 1) {
+			&sendmail::restart_sendmail();
+			}
+		elsif ($config{'mail_system'} == 0) {
+			&shutdown_mail_server();
+			&startup_mail_server();
+			}
+		elsif ($config{'mail_system'} == 2) {
+			&qmailadmin::restart_qmail();
+			}
+		}
+
+	# Update any virtusers with addresses in the old domain
+	foreach $v (&list_virtusers()) {
+		if ($v->{'from'} =~ /^(\S*)\@(\S+)$/ && $2 eq $_[1]->{'dom'}) {
+			$v->{'from'} = "$1\@$_[0]->{'dom'}";
+			&modify_virtuser($v, $v);
+			}
+		}
 	}
 }
 
@@ -238,8 +284,8 @@ return $cwfile;
 
 # list_virtusers()
 # Returns a list of a virtual mail address mappings. Each may actually have
-# an alias as its destination, and is automatically expanded to the destinations
-# for that alias.
+# an alias as its destination, and is automatically expanded to the
+# destinations for that alias.
 sub list_virtusers
 {
 # Build list of unix users, to exclude aliases with same name as users
@@ -433,9 +479,9 @@ elsif ($config{'mail_system'} == 0) {
 		local $alias = { "name" => $an,
 				 "enabled" => 1,
 				 "values" => \@to };
-		&sendmail::lock_alias_files($postfix_afiles);
+		&postfix::lock_alias_files($postfix_afiles);
 		&postfix::create_alias($alias, $postfix_afiles);
-		&sendmail::unlock_alias_files($postfix_afiles);
+		&postfix::unlock_alias_files($postfix_afiles);
 		&postfix::regenerate_aliases();
 		local $virt = { "name" => $_[1]->{'from'},
 				"value" => $an };
@@ -654,6 +700,35 @@ elsif ($config{'mail_system'} == 0) {
 	}
 }
 
+# create_mail_file(&user)
+# Creates a new empty mail file for a user, if necessary
+sub create_mail_file
+{
+&require_mail();
+local $mf;
+if ($config{'mail_system'} == 1) {
+	$mf = &sendmail::user_mail_file($_[0]->{'user'});
+	}
+elsif ($config{'mail_system'} == 0) {
+	# Find out from Postfix which file to delete
+	local ($s, $d) = &postfix::postfix_mail_system();
+	if ($s == 0) {
+		$mf = &postfix::postfix_mail_file($_[0]->{'user'});
+		}
+	}
+elsif ($config{'mail_system'} == 2) {
+	if ($qmailadmin::config{'mail_system'} == 0) {
+		$mf = &qmailadmin::user_mail_file($_[0]->{'user'});
+		}
+	}
+if ($mf && !-r $mf) {
+	# Now create it, owned by the user
+	open(MF, ">$mf");
+	close(MF);
+	chown($_[0]->{'uid'}, $_[0]->{'gid'}, $mf);
+	}
+}
+
 # delete_mail_file(&user)
 sub delete_mail_file
 {
@@ -663,7 +738,6 @@ if ($config{'mail_system'} == 1) {
 	unlink(&sendmail::user_mail_file($_[0]->{'user'}));
 	}
 elsif ($config{'mail_system'} == 0) {
-	# Find out from Postfix which file to delete
 	local $s = &postfix::postfix_mail_system();
 	unlink(&postfix::postfix_mail_file($_[0]->{'user'}));
 	}
@@ -846,12 +920,14 @@ sub check_alias_clash
 &require_mail();
 if ($config{'mail_system'} == 1) {
 	local @aliases = &sendmail::list_aliases($sendmail_afiles);
-	local ($clash) = grep { lc($_->{'name'}) eq lc($_[0]) } @aliases;
+	local ($clash) = grep { lc($_->{'name'}) eq lc($_[0]) &&
+				$_->{'enabled'} } @aliases;
 	return $clash;
 	}
 elsif ($config{'mail_system'} == 0) {
 	local @aliases = &postfix::list_aliases($postfix_afiles);
-	local ($clash) = grep { lc($_->{'name'}) eq lc($_[0]) } @aliases;
+	local ($clash) = grep { lc($_->{'name'}) eq lc($_[0]) &&
+				$_->{'enabled'} } @aliases;
 	return $clash;
 	}
 return undef;
@@ -1075,14 +1151,16 @@ return 1 if ($clash);
 if ($config{'mail_system'} == 1) {
 	local @aliases = &sendmail::list_aliases($sendmail_afiles);
 	local $an = $_[0] ? "$_[0]-$_[1]" : "default-$_[1]";
-	($clash) = grep { $_[0] && $_->{'name'} eq $_[0] ||
+	($clash) = grep { ($config{'alias_clash'} &&
+			   $_[0] && $_->{'name'} eq $_[0]) ||
 			  $_->{'name'} eq $an } @aliases;
 	return 2 if ($clash);
 	}
 elsif ($config{'mail_system'} == 0) {
 	local @aliases = &postfix::list_aliases($postfix_afiles);
 	local $an = $_[0] ? "$_[0]-$_[1]" : "default-$_[1]";
-	($clash) = grep { $_[0] && $_->{'name'} eq $_[0] ||
+	($clash) = grep { ($config{'alias_clash'} &&
+			   $_[0] && $_->{'name'} eq $_[0]) ||
 			  $_->{'name'} eq $an } @aliases;
 	return 2 if ($clash);
 	}
