@@ -8,15 +8,17 @@
 #	XXX test all the alias types
 #	XXX qmail not accepting mail! (works for users, not aliases?)
 #	XXX 'all mailboxes' support (use default)
+#
+# XXX multiple domain per user support
+#	XXX forwarding domain creation?
 
 do '../web-lib.pl';
 &init_config();
 %access = &get_module_acl();
-$master_admin = !$access{'noconfig'};
 $single_domain_mode = $access{'domains'} =~ /^\d+$/ &&
 		      !$access{'edit'} && !$access{'create'} &&
 		      !$access{'stop'} && !$access{'local'};
-if (!$master_admin) {
+if (!&master_admin()) {
 	# Allowed alias types are set by module config
 	%can_alias_types = map { $_, 1 } split(/,/, $config{'alias_types'});
 	}
@@ -65,11 +67,16 @@ closedir(DIR);
 return @rv;
 }
 
-# get_domain(id)
+# get_domain(id, [file])
+# Looks up a domain object by ID
 sub get_domain
 {
+if (defined($get_domain_cache{$_[0]})) {
+	return $get_domain_cache{$_[0]};
+	}
 local %dom;
-&read_file("$domains_dir/$_[0]", \%dom) || return undef;
+local $file = $_[1] || "$domains_dir/$_[0]";
+&read_file($file, \%dom) || return undef;
 $dom{'file'} = "$domains_dir/$_[0]";
 $dom{'id'} = $_[0];
 $dom{'mail'} = 1 if (!defined($dom{'mail'}));	# compat - assume mail is on
@@ -107,19 +114,33 @@ if (!defined($dom{'web_sslport'}) && $dom{'ssl'}) {
 	# compat - assume SSL port is current setting
 	$dom{'web_sslport'} = $web_sslport;
 	}
+if (!defined($dom{'prefix'})) {
+	# compat - assume that prefix is same as group
+	$dom{'prefix'} = $dom{'group'};
+	}
+if (!defined($dom{'home'})) {
+	local @u = getpwnam($dom{'user'});
+	$dom{'home'} = $u[7];
+	}
+$get_domain_cache{$_[0]} = \%dom;
 return \%dom;
 }
 
-# get_domain_by(field, value)
+# get_domain_by(field, value, [field, value, ...])
+# Looks up a domain by some field(s)
 sub get_domain_by
 {
 local $d;
+local @rv;
 foreach $d (&list_domains()) {
-	if ($d->{$_[0]} eq $_[1]) {
-		return $d;
+	local $i;
+	local $allok = 1;
+	for($i=0; $i<@_; $i+=2) {
+		$allok = 0 if ($d->{$_[$i]} ne $_[$i+1]);
 		}
+	push(@rv, $d) if ($allok);
 	}
-return undef;
+return wantarray ? @rv : $rv[0];
 }
 
 # domain_id()
@@ -150,7 +171,7 @@ unlink("$domains_dir/$_[0]->{'id'}");
 }
 
 # list_domain_users([&domain], [skipunix])
-# List all Unix users who are in the domain's primary group, except the owner.
+# List all Unix users who are in the domain's primary group.
 # If domain is omitted, returns local users.
 sub list_domain_users
 {
@@ -181,10 +202,19 @@ local @virts = &list_virtusers();
 
 local @users = &list_all_users_quotas();
 if ($_[0]) {
-	# Limit to domain users
+	# Limit to domain users.
 	@users = grep { $_->{'gid'} == $_[0]->{'gid'} ||
 			$_->{'user'} eq $_[0]->{'user'} } @users;
-	@users = grep { $_->{'user'} ne $_[0]->{'user'} } @users if ($_[1]);
+	if ($_[0]->{'parent'}) {
+		# This is a subdomain - exclude parent domain users
+		@users = grep { $_->{'home'} =~ /^$_[0]->{'home'}\// } @users;
+		}
+	elsif (&get_domain_by("parent", $_[0]->{'id'})) {
+		# This domain has subdomains - exclude their users
+		@users = grep { $_->{'home'} !~ /^$_[0]->{'home'}\/domains\// } @users;
+		}
+	@users = grep { $_->{'user'} ne $_[0]->{'user'} } @users
+		if ($_[1] || $_[0]->{'parent'});
 	}
 else {
 	# Limit to local users
@@ -336,7 +366,7 @@ if ($config{'generics'} && $firstemail) {
 	}
 }
 
-# modify_user(&user, &old, &domain)
+# modify_user(&user, &old, &domain, [noaliases])
 sub modify_user
 {
 &require_useradmin();
@@ -349,6 +379,7 @@ sub modify_user
 &useradmin::modify_user($_[1], $_[0]);
 &useradmin::unlock_user_files();
 &useradmin::made_changes();
+return if ($_[3]);		# no need to touch aliases and virtusers
 
 # Take away all virtusers and add new ones
 &delete_virtuser($_[1]->{'virt'}) if ($_[1]->{'virt'});
@@ -521,7 +552,7 @@ print "<center><font size=+1>",&text('indom', "<tt>$_[0]->{'dom'}</tt>"),
       "</font></center>\n";
 }
 
-# copy_skel_files(basedir, &user, [group])
+# copy_skel_files(basedir, &user, home, [group])
 # Copy files to the home directory of some new user
 sub copy_skel_files
 {
@@ -530,12 +561,12 @@ return if (!$uf);
 &require_useradmin();
 local $shell = $_[1]->{'shell'};
 $shell =~ s/^(.*)\///g;
-local $group = $_[2];
+local $group = $_[3];
 $group = getgrgid($_[1]->{'gid'}) if (!$group);
 $uf =~ s/\$group/$group/g;
 $uf =~ s/\$gid/$_[1]->{'gid'}/g;
 $uf =~ s/\$shell/$shell/g;
-&useradmin::copy_skel_files($uf, $_[1]->{'home'},
+&useradmin::copy_skel_files($uf, $_[2],
 			    $_[1]->{'uid'}, $_[1]->{'gid'});
 }
 
@@ -551,6 +582,14 @@ foreach $d (split(/\s+/, $access{'domains'})) {
 return 0;
 }
 
+# master_admin()
+# Returns 1 if the current user is the master Virtualmin admin, and can edit
+# global settings
+sub master_admin
+{
+return !$access{'noconfig'};
+}
+
 # domains_table(&domains)
 # Display a list of domains in a table, with links for editing
 sub domains_table
@@ -575,15 +614,20 @@ if ($config{'home_quotas'}) {
 	}
 print "</tr>\n";
 local $d;
-foreach $d (sort { $a->{'dom'} cmp $b->{'dom'} } @{$_[0]}) {
+local %done;
+foreach $d (sort { $a->{'user'} cmp $b->{'user'} ||
+		   $a->{'created'} <=> $b->{'created'} } @{$_[0]}) {
+	$done{$d->{'id'}}++;
 	print "<tr $cb>\n";
 	local $dn = $d->{'disabled'} ? "<i>$d->{'dom'}</i>" : $d->{'dom'};
+	local $pfx = $d->{'parent'} && $done{$d->{'parent'}} ?
+		"&nbsp;&nbsp;" : "";
 	if ($access{'edit'}) {
-		print "<td><a href='edit_domain.cgi?dom=$d->{'id'}'>",
+		print "<td>$pfx<a href='edit_domain.cgi?dom=$d->{'id'}'>",
 		      "$dn</a></td>\n";
 		}
 	else {
-		print "<td>$dn</td>\n";
+		print "<td>$pfx$dn</td>\n";
 		}
 	print "<td>$d->{'user'}</td>\n";
 	print "<td>$d->{'owner'}</td>\n";
@@ -596,7 +640,7 @@ foreach $d (sort { $a->{'dom'} cmp $b->{'dom'} } @{$_[0]}) {
 	if ($config{'mail'}) {
 		if ($d->{'mail'}) {
 			local @aliases = &list_domain_aliases($d);
-			printf "<td>%d%s&nbsp;(<a href='list_users.cgi?dom=$d->{'id'}'>$text{'index_list'}</a>)</td>\n", scalar(@users), $d->{'mailboxlimit'} ? " of $d->{'mailboxlimit'}" : "";
+			printf "<td>%d&nbsp;(<a href='list_users.cgi?dom=$d->{'id'}'>$text{'index_list'}</a>)</td>\n", scalar(@users);
 			printf "<td>%d&nbsp;(<a href='list_aliases.cgi?dom=$d->{'id'}'>$text{'index_list'}</a>)</td>\n", scalar(@aliases);
 			}
 		else {
@@ -604,9 +648,19 @@ foreach $d (sort { $a->{'dom'} cmp $b->{'dom'} } @{$_[0]}) {
 			}
 		}
 	if ($config{'home_quotas'}) {
-		print "<td>",$d->{'quota'} ?
-			&quota_show($d->{'quota'}, $config{'home_quotas'}) :
-			$text{'form_unlimit'},"</td>\n";
+		if ($d->{'parent'}) {
+			if ($done{$d->{'parent'}}) {
+				print "<td>&nbsp;&nbsp;\"</td>\n";
+				}
+			else {
+				print "<td>$text{'index_samequ'}</td>\n";
+				}
+			}
+		else {
+			print "<td>",$d->{'quota'} ?
+			  &quota_show($d->{'quota'}, $config{'home_quotas'}) :
+			  $text{'form_unlimit'},"</td>\n";
+			}
 		local $ut = $duser->{'uquota'} + $duser->{'umquota'};
 		foreach $u (@users) {
 			$ut += $u->{'uquota'} + $u->{'umquota'}
@@ -620,26 +674,26 @@ print "</table>\n";
 }
 
 # userdom_name(name, &domain)
-# Returns a username with the domain group appended somehow
+# Returns a username with the domain prefix (usually group) appended somehow
 sub userdom_name
 {
 if ($config{'append_style'} == 0) {
-	return $_[0].".".$_[1]->{'group'};
+	return $_[0].".".$_[1]->{'prefix'};
 	}
 elsif ($config{'append_style'} == 1) {
-	return $_[0]."-".$_[1]->{'group'};
+	return $_[0]."-".$_[1]->{'prefix'};
 	}
 elsif ($config{'append_style'} == 2) {
-	return $_[1]->{'group'}.".".$_[0];
+	return $_[1]->{'prefix'}.".".$_[0];
 	}
 elsif ($config{'append_style'} == 3) {
-	return $_[1]->{'group'}."-".$_[0];
+	return $_[1]->{'prefix'}."-".$_[0];
 	}
 elsif ($config{'append_style'} == 4) {
-	return $_[0]."_".$_[1]->{'group'};
+	return $_[0]."_".$_[1]->{'prefix'};
 	}
 elsif ($config{'append_style'} == 5) {
-	return $_[1]->{'group'}."_".$_[0];
+	return $_[1]->{'prefix'}."_".$_[0];
 	}
 else {
 	&error("Unknown append_style $config{'append_style'}!");
@@ -647,10 +701,10 @@ else {
 }
 
 # remove_userdom(name, &domain)
-# Returns a username with the domain group stripped off
+# Returns a username with the domain prefix (group) stripped off
 sub remove_userdom
 {
-local $g = $_[1]->{'group'};
+local $g = $_[1]->{'prefix'};
 local $rv = $_[0];
 ($rv =~ s/(\.|\-|_)\Q$g\E$//) || ($rv =~ s/^\Q$g\E(\.|\-|_)//);
 return $rv;
@@ -728,11 +782,13 @@ sub null_print { }
 # (0=failed, 1=success) and an optional message. Also outputs status messages.
 sub send_domain_email
 {
-&ensure_template("domain-template");
-return (1, undef) if ($config{'domain_template'} eq 'none');
+local $file = $_[0]->{'parent'} ? "subdomain-template" : "domain-template";
+local $cfg = $_[0]->{'parent'} ? "subdomain_template" : "domain_template";
+&ensure_template($file);
+return (1, undef) if ($config{$cfg} eq 'none');
 &$first_print($text{'setup_email'});
-local $tmpl = $config{'domain_template'} eq 'default' ?
-	"$module_config_directory/domain-template" : $config{'domain_template'};
+local $tmpl = $config{$cfg} eq 'default' ?
+	"$module_config_directory/$file" : $config{$cfg};
 local @erv = &send_template_email($tmpl,
 				  $_[0]->{'email'} ||
 				    $_[0]->{'user'}.'@'.&get_system_hostname(),
@@ -986,7 +1042,7 @@ for($i=0; $i<=@values+2; $i++) {
 		}
 	print "</select>\n";
 	print "<input name=val_$i size=30 value=\"$val\">\n";
-	if ($config{'edit_afiles'} || $master_admin) {
+	if ($config{'edit_afiles'} || &master_admin()) {
 		local $prog = $type == 2 ? "edit_afile.cgi" :
 			      $type == 5 ? "edit_rfile.cgi" :
 			      $type == 6 ? "edit_ffile.cgi" : undef;
@@ -1129,11 +1185,16 @@ while($_[0]->{$gid}) {
 return $gid;
 }
 
-# server_home_directory(&domain)
+# server_home_directory(&domain, [&parentdomain])
 # Returns the home directory for a new virtual server user
 sub server_home_directory
 {
-if ($config{'home_format'}) {
+&require_useradmin();
+if ($_[0]->{'parent'}) {
+	# Owned by some existing user, so under his home
+	return "$_[1]->{'home'}/domains/$_[0]->{'dom'}";
+	}
+elsif ($config{'home_format'}) {
 	# Use the template from the module config
 	local $home = "$home_base/$config{'home_format'}";
 	return &substitute_template($home, $_[0]);
@@ -1215,7 +1276,7 @@ foreach $u (@{$_[0]}) {
 	      "user=$u->{'user'}'>",
 	      ($domuser ? "<b>$pop3</b>" : $pop3),"</a></td>\n";
 	print "<td>$u->{'user'}</td>\n";
-	print "<td>$u->{'real'}</td>\n";
+	print "<td>",($u->{'real'} || "<br>"),"</td>\n";
 	local $quota;
 	$quota += $u->{'quota'} if ($config{'home_quotas'});
 	$quota += $u->{'mquota'} if ($config{'mail_quotas'} &&
@@ -1292,7 +1353,7 @@ sub setup_virt
 {
 &foreign_require("net", "net-lib.pl");
 local @boot = &net::active_interfaces();
-&$first_print($text{'setup_virt'});
+&$first_print(&text('setup_virt', $_[0]->{'ip'}));
 local ($iface) = grep { $_->{'fullname'} eq $config{'iface'} } @boot;
 local $b;
 local $vmax = int($net::min_virtual_number);
@@ -1588,6 +1649,34 @@ if ($ok) {
 	local $d;
 	$no_restart_webmin = 1;
 	DOMAIN: foreach $d (@{$_[1]}) {
+		if ($d->{'missing'}) {
+			# This domain doesn't exist yet - need to re-create it
+			&$first_print(&text('restore_createdomain',
+				      $d->{'dom'}));
+			$d = &get_domain(undef,
+				"$restoredir/$d->{'dom'}_virtualmin");
+			local $cerr = &virtual_server_clashes($d);
+			if ($cerr) {
+				&$second_print(&text('restore_eclash', $cerr));
+				$ok = 0;
+				last DOMAIN;
+				}
+			local ($parentdom, $parentuser);
+			if ($d->{'parent'}) {
+				# Does the parent exist?
+				$parentdom = &get_domain($d->{'parent'});
+				if (!$parentdom) {
+					&$second_print($text{'restore_epar'});
+					$ok = 0;
+					last DOMAIN;
+					}
+				$parentuser = $parentdom->{'user'};
+				}
+			&$indent_print();
+			&create_virtual_server($d);
+			&$outdent_print();
+			}
+
 		&$first_print(&text('restore_fordomain', $d->{'dom'}));
 		#&$second_print();
 		&$indent_print();
@@ -2079,5 +2168,204 @@ if ($? || $out =~ /permission\s+denied/i) {
 	}
 }
 
+# free_ip_address()
+# Returns an IP address within the allocation range which is not currently used
+sub free_ip_address
+{
+&foreign_require("net", "net-lib.pl");
+local %taken = map { $_->{'address'}, $_ } (&net::boot_interfaces(),
+					    &net::active_interfaces());
+local @ranges = split(/\s+/, $config{'ip_ranges'});
+local $r;
+foreach $r (@ranges) {
+	$r =~ /^(\d+\.\d+\.\d+)\.(\d+)\-(\d+)$/ || next;
+	local ($base, $s, $e) = ($1, $2, $3);
+	local $j;
+	for($j=$s; $j<=$e; $j++) {
+		local $try = "$base.$j";
+		return $try if (!$taken{$try});
+		}
+	}
+return undef;
+}
+
+# setup_for_subdomain(&parent-domain, subdomain-user, &sub-domain)
+# Ensures that this virtual server can host sub-domains
+sub setup_for_subdomain
+{
+&system_logged("mkdir '$_[0]->{'home'}/domains' 2>/dev/null");
+&system_logged("chmod 755 '$_[0]->{'home'}/domains'");
+&system_logged("chown $_[0]->{'uid'}:$_[0]->{'gid'} '$_[0]->{'home'}/domains'");
+}
+
+# count_domains(&parent)
+# Returns the number of domains owned by this domain, and the max allowed
+# for the current user
+sub count_domains
+{
+local $count = 1;
+local @subs = &get_domain_by("parent", $_[0]->{'id'});
+$count += @subs;
+local $max;
+if (&master_admin() || $_[0]->{'domslimit'} eq "*") {
+	$max = 99999999;	# a huge number
+	}
+else {
+	$max = $_[0]->{'domslimit'};
+	}
+return ($count, $max);
+}
+
+# count_mailboxes(&parent)
+# Returns the number of mailboxes in this domain and all subdomains, and the
+# max allowed for the current user
+sub count_mailboxes
+{
+local $count = 0;
+local $doms = 0;
+local $parent = $_[0]->{'parent'} ? &get_domain($_[0]->{'parent'}) : $_[0];
+local $d;
+foreach $d ($parent, &get_domain_by("parent", $parent->{'id'})) {
+	local @users = &list_domain_users($d);
+	$count += @users;
+	$doms++;
+	}
+return ( $count, $parent->{'mailboxlimit'} ? $parent->{'mailboxlimit'} : 0,
+	 $doms );
+}
+
+# database_name(&domain)
+# Returns a suitable database name for a domain
+sub database_name
+{
+local $db = $_[0]->{'prefix'};
+$db =~ s/[\.\-]/_/g;
+return $db;
+}
+
+# unixuser_name(domainname)
+# Returns a Unix username for soem domain, or undef if none can be found
+sub unixuser_name
+{
+$_[0] =~ /^([^\.]+)/;
+local ($try1, $user) = ($1, $1);
+if (defined(getpwnam($try1)) || $config{'longname'}) {
+	$user = $_[0];
+	$try2 = $user;
+	if (defined(getpwnam($try))) {
+		return (undef, $try1, $try2);
+		&error(&text('setup_eauto', $try1, $try2));
+		}
+	}
+return ($user);
+}
+
+# virtual_server_clashes(&dom)
+# Returns a clash error message if any were found for some new domain
+sub virtual_server_clashes
+{
+local ($dom) = @_;
+my $f;
+foreach $f (@features) {
+	if ($dom->{$f}) {
+		local $cfunc = "check_${f}_clash";
+		if (&$cfunc($dom)) {
+			return &text('setup_e'.$f, $dom->{'dom'}, $dom->{'db'},
+				     $dom->{'user'}, $dom->{'group'});
+			}
+		}
+	}
+return undef;
+}
+
+# create_virtual_server(&domain, [&parent-domain], [parent-user])
+# Given a complete domain object, setup all it's features
+sub create_virtual_server
+{
+local ($dom, $parentdom, $parentuser) = @_;
+
+# Run the before command
+&set_domain_envs($_[0], "CREATE_DOMAIN");
+local $merr = &making_changes();
+return &text('setup_emaking', "<tt>$merr</tt>") if (defined($merr));
+
+# Get ready for hosting a subdomain
+if ($dom->{'parent'}) {
+	&setup_for_subdomain($parentdom, $parentuser, $dom);
+	}
+
+# Set up all the selected features
+my $f;
+foreach $f (@features) {
+	if ($dom->{$f}) {
+		local $sfunc = "setup_$f";
+		&$sfunc($dom);
+		}
+	}
+&run_post_actions();
+
+# Add virtual IP address, if needed
+if ($dom->{'virt'}) {
+	&setup_virt($dom);
+	}
+
+# Add a virtuser for the unix user, if requested
+if ($in{'mailbox'}) {
+	&$first_print($text{'setup_mailbox'});
+	local $virt = { 'from' => $user."\@".$dom->{'dom'},
+			'to' => [ $user ] };
+	&create_virtuser($virt);
+	&$second_print($text{'setup_done'});
+	}
+
+# Save domain details
+&$first_print($text{'setup_save'});
+&save_domain($dom);
+&$second_print($text{'setup_done'});
+
+# Notify the owner via email
+&send_domain_email($dom);
+
+# Update the parent domain Webmin user
+if ($parentdom) {
+	&modify_webmin($parentdom, $parentdom);
+	}
+
+if ($remote_user) {
+	# Add to this user's list of domains if needed
+	local %access = &get_module_acl();
+	if (!&can_edit_domain($dom)) {
+		$access{'domains'} = join(" ", split(/\s+/, $access{'domains'}),
+					       $dom->{'id'});
+		&save_module_acl(\%access);
+		}
+	}
+
+# Run the after creation command
+&made_changes();
+
+return undef;
+}
+
+# register_post_action(&function, args)
+sub register_post_action
+{
+push(@post_actions, [ @_ ]);
+}
+
+# run_post_actions()
+# Run all registered post-modification actions
+sub run_post_actions
+{
+local $a;
+local %done;
+foreach $a (@post_actions) {
+	next if ($done{join(",", @$a)}++);
+	local ($afunc, @aargs) = @$a;
+	&$afunc(@aargs);
+	}
+@post_actions = ( );
+}
+
 1;
 
